# Django Stripe Stuff
from django_stripe.actions import StripeSubscriptionAction
from django_stripe.webhooks.register import StripeWebhook


class CustomerSubscriptionStripeWebhook(StripeWebhook):
    def process_webhook(self):
        if self.event.validated_message:
            StripeSubscriptionAction().sync(
                self.event.validated_message["data"]["object"].copy(),
            )


class CustomerSubscriptionCreatedWebhook(CustomerSubscriptionStripeWebhook):
    name = "customer.subscription.created"
    description = (
        "Occurs whenever a customer with no subscription is signed up for a plan."
    )


class CustomerSubscriptionDeletedWebhook(CustomerSubscriptionStripeWebhook):
    name = "customer.subscription.deleted"
    description = "Occurs whenever a customer ends their subscription."

    def process_webhook(self):
        if self.event.validated_message:
            StripeSubscriptionAction().soft_delete(
                self.event.validated_message["data"]["object"]["id"]
            )


class CustomerSubscriptionTrialWillEndWebhook(CustomerSubscriptionStripeWebhook):
    name = "customer.subscription.trial_will_end"
    description = (
        "Occurs three days before the trial "
        "period of a subscription is scheduled to end."
    )


class CustomerSubscriptionUpdatedWebhook(CustomerSubscriptionStripeWebhook):
    name = "customer.subscription.updated"
    description = (
        "Occurs whenever a subscription changes. "
        "Examples would include switching from one plan to another, "
        "or switching status from trial to active."
    )



from django_stripe.models.abstracts.billings import AbstractStripeSubscription
from django_stripe.models.abstracts.core import (
    AbstractStripeCustomer,
    AbstractStripeEvent,
)
from django_stripe.models.abstracts.payment_methods import AbstractStripeCard
from django_stripe.models.abstracts.products import (
    AbstractStripeCoupon,
    AbstractStripePrice,
    AbstractStripeProduct,
)

__all__ = [
    "AbstractStripeSubscription",
    "AbstractStripeCustomer",
    "AbstractStripeEvent",
    "AbstractStripeCard",
    "AbstractStripeProduct",
    "AbstractStripePrice",
    "AbstractStripeCoupon",
]



# Django Stripe Stuff
from django_stripe.settings import stripe_settings


class StripeRouter:
    def allow_migrate(self, db, app_label, model_name=None, **hints):
        # Allow migration only for the 'StripeCustomer' model
        if app_label == "django_stripe" and (
            "__all__" in stripe_settings.ALLOWED_MODELS
            or model_name in stripe_settings.ALLOWED_MODELS
        ):
            return True
        return False



# Third Party Stuff
from django.db import models

# Django Stripe Stuff
from django_stripe.models.abstracts.mixins import AbstractStripeModel
from django_stripe.utils import CURRENCY_SYMBOLS, Currency


class AbstractStripeCoupon(AbstractStripeModel):
    """
    A coupon contains information about a percent-off or
    amount-off discount we might want to apply to a customer.

    Coupons may be applied to invoices or orders.
    Coupons do not work with conventional one-off charges.

    Stripe documentation: https://stripe.com/docs/api/coupons
    """

    ONCE = "once"
    REPEATING = "repeating"
    FOREVER = "forever"

    STRIPE_COUPON_DURATION_TYPES = (
        (ONCE, "Once"),
        (REPEATING, "Repeating"),
        (FOREVER, "Forever"),
    )

    name = models.CharField(
        max_length=64,
        blank=True,
        help_text=(
            "Name of the coupon displayed to customers "
            "on for instance invoices or receipts"
        ),
    )
    applies_to = models.JSONField(
        null=True,
        blank=True,
        help_text=(
            "Contains information about what product this coupon applies to. "
            "This field is not included by default. "
            "To include it in the response, expand the applies_to field"
        ),
    )
    amount_off = models.DecimalField(
        decimal_places=2,
        max_digits=9,
        null=True,
        blank=True,
        help_text=(
            "Amount (in the currency specified) "
            "that will be taken off the subtotal "
            "of any invoices for this customer"
        ),
    )
    currency = models.CharField(
        choices=Currency.choices,
        default=Currency.USD.value,
        max_length=3,
        help_text=(
            "If amount_off has been set, the three-letter ISO code "
            "for the currency of the amount to take off"
        ),
    )
    duration = models.CharField(
        choices=STRIPE_COUPON_DURATION_TYPES,
        max_length=16,
        default="once",
        help_text=(
            "One of forever, once, and repeating. "
            "Describes how long a customer who applies this coupon "
            "will get the discount"
        ),
    )
    duration_in_months = models.PositiveIntegerField(
        null=True,
        blank=True,
        help_text=(
            "Required only if duration is repeating, "
            "in which case it must be a positive integer that "
            "specifies the number of months the discount will be in effect"
        ),
    )
    max_redemptions = models.PositiveIntegerField(
        null=True,
        blank=True,
        help_text="A positive integer specifying the number of times the coupon can "
        "be redeemed before it’s no longer valid",
    )
    percent_off = models.FloatField(
        null=True,
        blank=True,
        help_text="Percent that will be taken off the subtotal of any invoices "
        "for this customer for the duration of the coupon",
    )
    redeem_by = models.DateTimeField(
        null=True,
        blank=True,
        help_text="Date after which the coupon can no longer be redeemed",
    )
    times_redeemed = models.PositiveIntegerField(
        null=True,
        blank=True,
        help_text="Number of times this coupon has been applied to a customer",
    )

    valid = models.BooleanField(default=False)

    def __str__(self):
        if self.amount_off is None:
            description = "{}% off".format(
                self.percent_off,
            )
        else:
            description = "{}{}".format(
                CURRENCY_SYMBOLS.get(self.currency, ""), self.amount_off
            )

        return "Coupon for {}, {}".format(description, self.duration)

    class Meta:
        abstract = True



# Third Party Stuff
import stripe

# Django Stripe Stuff
from django_stripe.actions.mixins import (
    StripeSoftDeleteActionMixin,
    StripeSyncActionMixin,
)
from django_stripe.models import StripeCoupon, StripePrice, StripeProduct


class StripeProductAction(StripeSyncActionMixin, StripeSoftDeleteActionMixin):
    """
    Actions related to products in Stripe that can be used for various purposes.

    Syncing is done by retrieving a batch of products from the Stripe API
    and then iterating over them and calling sync method on each of them.

    Example:
        from django_stripe.actions import StripeProductAction

        StripeProductAction.sync_all()
    """

    model_class = StripeProduct
    stripe_object_class = stripe.Product


class StripePriceAction(StripeSyncActionMixin, StripeSoftDeleteActionMixin):
    product_model_class = StripeProduct
    stripe_product_class = stripe.Product
    model_class = StripePrice
    stripe_object_class = stripe.Product

    def pre_set_defualt(self, stripe_data: dict):
        """
        Sync product if not exist.
        Update product field with product object
        Args:
            stripe_data: data from Stripe API representing a price
        """
        product = self.product_model_class.objects.filter(
            stripe_id=stripe_data["product"]
        ).first()

        if not product:
            stripe_product = self.stripe_product_class.retrieve(stripe_data["product"])
            product = StripeProductAction().sync(stripe_product)

        stripe_data["product"] = product


class StripeCouponAction(StripeSoftDeleteActionMixin, StripeSyncActionMixin):
    model_class = StripeCoupon
    stripe_object_class = stripe.Coupon



from django_stripe.utils.currencies import Currency
from django_stripe.utils.utils import (
    CURRENCY_SYMBOLS,
    ZERO_DECIMAL_CURRENCIES,
    convert_amount_for_db,
    convert_epoch,
)

__all__ = [
    "Currency",
    "convert_epoch",
    "convert_amount_for_db",
    "ZERO_DECIMAL_CURRENCIES",
    "CURRENCY_SYMBOLS",
]



from django_stripe.webhooks.billings.subscriptions import (
    CustomerSubscriptionCreatedWebhook,
    CustomerSubscriptionDeletedWebhook,
    CustomerSubscriptionTrialWillEndWebhook,
    CustomerSubscriptionUpdatedWebhook,
)

__all__ = (
    "CustomerSubscriptionCreatedWebhook",
    "CustomerSubscriptionUpdatedWebhook",
    "CustomerSubscriptionDeletedWebhook",
    "CustomerSubscriptionTrialWillEndWebhook",
)



# Third Party Stuff
import stripe

# Django Stripe Stuff
from django_stripe.actions.mixins import StripeSyncActionMixin
from django_stripe.models import StripeCard


class StripeCardAction(StripeSyncActionMixin):
    model_class = StripeCard

    def __init__(self, customer):
        """
        Args:
            customer: django-stripe Customer
        """
        self.customer = customer

    def pre_set_defualt(self, stripe_data: dict):
        """
        Add customer data to stripe_data
        """
        stripe_data["customer"] = self.customer.stripe_id

    def delete(self, stripe_id):
        """
        Deletes a card from a customer
        Args:
            source_stripe_id: the Stripe ID of the payment source to delete
        Ref Docs: https://stripe.com/docs/api/cards/delete
        """
        stripe.Customer.delete_source(self.customer.stripe_id, stripe_id)
        return self.model_class.objects.filter(stripe_id=stripe_id).delete()



# Third Party Stuff
from django.db import models

# Django Stripe Stuff
from django_stripe.models.abstracts.core import AbstractStripeEvent


class StripeEvent(AbstractStripeEvent):
    customer = models.ForeignKey(
        "django_stripe.StripeCustomer",
        to_field="stripe_id",
        on_delete=models.SET_NULL,
        null=True,
        related_name="events",
    )



# Standard Library
import decimal
from datetime import datetime

CURRENCY_SYMBOLS = {
    "aud": "\u0024",
    "cad": "\u0024",
    "chf": "\u0043\u0048\u0046",
    "cny": "\u00a5",
    "eur": "\u20ac",
    "gbp": "\u00a3",
    "jpy": "\u00a5",
    "myr": "\u0052\u004d",
    "sgd": "\u0024",
    "usd": "\u0024",
}

# currencies those amount=1 means 100 cents
# https://support.stripe.com/questions/which-zero-decimal-currencies-does-stripe-support
ZERO_DECIMAL_CURRENCIES = [
    "bif",
    "clp",
    "djf",
    "gnf",
    "jpy",
    "kmf",
    "krw",
    "mga",
    "pyg",
    "rwf",
    "vuv",
    "xaf",
    "xof",
    "xpf",
]


def convert_epoch(epoch):
    """
    Convert epoch to datetime
    """
    return str(datetime.utcfromtimestamp(epoch))


def convert_amount_for_db(amount, currency="usd"):
    return (
        (amount / decimal.Decimal("100"))
        if currency.lower() not in ZERO_DECIMAL_CURRENCIES
        else decimal.Decimal(amount)
    )



# Standard Library
import uuid

# Third Party Stuff
from django.db import models


class UUIDModel(models.Model):
    """An abstract base class model that makes primary key `id` as UUID
    instead of default auto incremented number.
    """

    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4)

    class Meta:
        abstract = True


class TimeStampedUUIDModel(UUIDModel):
    """An abstract base class model that provides self-updating
    ``created`` and ``modified`` fields with UUID as primary_key field.
    """

    created_at = models.DateTimeField(auto_now_add=True, editable=False)
    updated_at = models.DateTimeField(auto_now=True, editable=False)

    class Meta:
        abstract = True


class AbstractStripeModel(TimeStampedUUIDModel):
    """
    An abstract base class model that provides stripe_id field
    with UUID as primary_key along with self-updating
    created_at and modified_at fields
    """

    stripe_id = models.CharField(max_length=255, unique=True)
    livemode = models.BooleanField(
        default=False,
        help_text=(
            "Has the value true if the object exists in live mode or"
            "the value false if the object exists in test mode."
        ),
    )
    metadata = models.JSONField(
        null=True,
        blank=True,
        default=dict,
        help_text=(
            "Set of key-value pairs that you can attach to an object."
            "This can be useful for storing additional information "
            "about the object in a structured format."
        ),
    )

    # Soft delete product in DB on deletion from stripe
    date_purged = models.DateTimeField(null=True, editable=False)

    class Meta:
        abstract = True



from django_stripe.models.abstracts.products.coupons import AbstractStripeCoupon
from django_stripe.models.abstracts.products.prices import AbstractStripePrice
from django_stripe.models.abstracts.products.products import AbstractStripeProduct

__all__ = ("AbstractStripeProduct", "AbstractStripePrice", "AbstractStripeCoupon")



# Third Party Stuff
from django.db import models

# Django Stripe Stuff
from django_stripe.models.abstracts.mixins import AbstractStripeModel
from django_stripe.utils import Currency

CHARGE_AUTOMATICALLY = "charge_automatically"
SEND_INVOICE = "send_invoice"

INVOICE_COLLECTION_METHOD_TYPES = (
    (CHARGE_AUTOMATICALLY, "Charge Automatically"),
    (SEND_INVOICE, "Send_Invoice"),
)


class AbstractStripeSubscription(AbstractStripeModel):
    """
    Subscriptions allow us to charge a customer on a recurring basis.
    Stripe documentation: https://stripe.com/docs/api/subscriptions
    """

    # https://stripe.com/docs/api/subscriptions/object#subscription_object-status
    INCOMPLETE = "incomplete"  # if the initial payment attempt fails

    # If the first invoice is not paid within 23 hours
    # (Its a terminal status)
    INCOMPLETE_EXPIRED = "incomplete_expired"

    TRIALING = "trialing"
    ACTIVE = "active"
    # It becomes past_due when payment to renew it fails
    PAST_DUE = "past_due"

    # 1. It becomes canceled when failed payment is not paid
    #    after all retries / by the due date
    # 2. Can be cancelled manually (its a terminal status)
    CANCELED = "canceled"

    # It becomes unpaid when failed payment
    # is not paid after all retries / by the due date
    UNPAID = "unpaid"

    STATUS_CURRENT = [TRIALING, ACTIVE]
    STATUS_CANCELLED = [CANCELED, UNPAID]

    SUBSCRIPTION_STATUS_TYPES = (
        (INCOMPLETE, "Incomplete"),
        (INCOMPLETE_EXPIRED, "Incomplete Expired"),
        (TRIALING, "Trialing"),
        (ACTIVE, "Active"),
        (PAST_DUE, "Past Due"),
        (CANCELED, "Canceled"),
        (UNPAID, "Unpaid"),
    )

    items = models.JSONField(
        null=True,
        blank=True,
        help_text="List of subscription items, each with an attached price.",
    )
    application_fee_percent = models.DecimalField(
        decimal_places=2,
        max_digits=5,
        null=True,
        blank=True,
        help_text=(
            "A positive decimal that represents the fee percentage of the "
            "subscription invoice amount that will be transferred to the application "
            "owner's Stripe account each billing period."
        ),
    )
    automatic_tax = models.JSONField(
        null=True,
        blank=True,
        help_text="Automatic tax settings for this subscription.",
    )
    billing_cycle_anchor = models.DateTimeField(
        null=True,
        blank=True,
        help_text=(
            "Determines the date of the first full invoice, and, for plans "
            "with `month` or `year` intervals, the day of the month for subsequent "
            "invoices"
        ),
    )
    billing_thresholds = models.JSONField(
        null=True,
        blank=True,
        help_text=(
            "Define thresholds at which an invoice will be sent, "
            "and the subscription advanced to a new billing period"
        ),
    )
    cancel_at = models.DateTimeField(
        null=True,
        blank=True,
        help_text=(
            "A date in the future at which the subscription will automatically "
            "get canceled"
        ),
    )
    cancel_at_period_end = models.BooleanField(
        default=False,
        help_text=(
            "If the subscription has been canceled with the ``at_period_end`` "
            "flag set to true, ``cancel_at_period_end`` "
            "on the subscription will be true. "
            "We can use this attribute to determine whether a subscription that has a "
            "status of active is scheduled to be canceled at the end of the "
            "current period"
        ),
    )
    canceled_at = models.DateTimeField(
        null=True,
        blank=True,
        help_text=(
            "If the subscription has been canceled, the date of that cancellation. "
            "If the subscription was canceled with ``cancel_at_period_end``, "
            "canceled_at will still reflect the date "
            "of the initial cancellation request, "
            "not the end of the subscription period "
            "when the subscription is automatically "
            "moved to a canceled state"
        ),
    )
    cancellation_details = models.JSONField(
        null=True,
        blank=True,
        help_text=("Details about why this subscription was cancelled"),
    )
    collection_method = models.CharField(
        choices=INVOICE_COLLECTION_METHOD_TYPES,
        max_length=32,
        help_text=(
            "Either `charge_automatically`, or `send_invoice`. "
            "When charging automatically, "
            "Stripe will attempt to pay this subscription "
            "at the end of the cycle using "
            "the default source attached to the customer. "
            "When sending an invoice, Stripe will email us customer an invoice with "
            "payment instructions"
        ),
    )
    current_period_end = models.DateTimeField(
        help_text="End of the current period for which the subscription has been "
        "invoiced. At the end of this period, a new invoice will be created"
    )
    current_period_start = models.DateTimeField(
        help_text=(
            "Start of the current period for which the subscription has "
            "been invoiced"
        )
    )
    days_until_due = models.IntegerField(
        null=True,
        blank=True,
        help_text=(
            "Number of days a customer has to pay invoices generated by this "
            "subscription. This value will be `null` for subscriptions where "
            "`billing=charge_automatically`"
        ),
    )
    default_payment_method = models.TextField(blank=True)
    default_source = models.TextField(blank=True)

    default_tax_rates = models.JSONField(null=True, blank=True)
    discount = models.JSONField(null=True, blank=True)
    ended_at = models.DateTimeField(
        null=True,
        blank=True,
        help_text=(
            "If the subscription has ended (either because it was canceled or "
            "because the customer was switched to a subscription to a new plan), "
            "the date the subscription ended"
        ),
    )
    next_pending_invoice_item_invoice = models.DateTimeField(
        null=True,
        blank=True,
        help_text=(
            "Specifies the approximate timestamp on which any pending "
            "invoice items will be billed according to the schedule provided at "
            "pending_invoice_item_interval"
        ),
    )
    pause_collection = models.JSONField(
        null=True,
        blank=True,
        help_text=(
            "If specified, payment collection for this subscription will be paused."
        ),
    )
    pending_invoice_item_interval = models.JSONField(
        null=True,
        blank=True,
        help_text=(
            "Specifies an interval for how often to bill for any "
            "pending invoice items. It is analogous to calling Create an invoice "
            "for the given subscription at the specified interval"
        ),
    )
    pending_setup_intent = models.TextField(blank=True)
    pending_update = models.JSONField(
        null=True,
        blank=True,
        help_text=(
            "If specified, pending updates that will be applied to the "
            "subscription once the latest_invoice has been paid"
        ),
    )
    quantity = models.IntegerField(
        null=True,
        blank=True,
        help_text=(
            "The quantity applied to this subscription. This value will be "
            "`null` for multi-plan subscriptions"
        ),
    )
    start_date = models.DateTimeField(
        null=True,
        blank=True,
        help_text=(
            "Date when the subscription was first created. The date "
            "might differ from the created date due to backdating"
        ),
    )
    status = models.CharField(
        choices=SUBSCRIPTION_STATUS_TYPES,
        max_length=32,
        help_text="The status of this subscription",
    )

    # tax rate
    trial_end = models.DateTimeField(
        null=True,
        blank=True,
        help_text="If the subscription has a trial, the end of that trial",
    )
    trial_start = models.DateTimeField(
        null=True,
        blank=True,
        help_text="If the subscription has a trial, the beginning of that trial",
    )
    trial_settings = models.JSONField(
        null=True,
        blank=True,
        help_text="Settings related to subscription trials.",
    )
    latest_invoice = models.CharField(
        max_length=255,
        blank=True,
        help_text="The most recent invoice this subscription has generated.",
    )
    currency = models.CharField(
        choices=Currency.choices,
        default=Currency.USD.value,
        max_length=3,
        help_text=(
            "The currency the customer can be charged "
            "in for recurring billing purposes"
        ),
    )

    class Meta:
        abstract = True



# Third Party Stuff
from django.contrib.postgres.fields import ArrayField
from django.db import models

# Django Stripe Stuff
from django_stripe.models.abstracts.mixins import AbstractStripeModel


class AbstractStripeProduct(AbstractStripeModel):
    """
    Products describe the specific goods or services you offer to your customers.
    For example, you might offer a Standard and
    Premium version of your goods or service;
    each version would be a separate Product.
    They can be used in conjunction with Prices
    to configure pricing in Payment Links, Checkout, and Subscriptions.
    Stripe documentation: https://stripe.com/docs/api/products
    """

    active = models.BooleanField(
        help_text=("Whether the product is currently available for purchase."),
    )
    description = models.TextField(
        null=True,
        help_text=(
            "The product’s description, meant to be displayable to the customer. "
            "Use this field to optionally store a long form explanation "
            "of the product being sold for your own rendering purposes."
        ),
    )
    name = models.CharField(
        max_length=255,
        help_text=(
            "The product’s name, meant to be displayable to the customer. "
            "Whenever this product is sold via a subscription, "
            "name will show up on associated invoice line item descriptions."
        ),
    )

    statement_descriptor = models.TextField(
        null=True,
        help_text=(
            "Extra information about a product which will appear "
            "on your customer’s credit card statement."
            "In the case that multiple products are billed at once, "
            "the first statement descriptor will be used."
        ),
    )
    tax_code = models.CharField(max_length=255, null=True, help_text="A tax code ID.")
    unit_label = models.CharField(
        max_length=255,
        null=True,
        help_text=(
            "A label that represents units of this product in Stripe"
            " and on customers’ receipts and invoices."
            "When set, this will be included in associated "
            "invoice line item descriptions."
        ),
    )
    images = ArrayField(
        models.CharField(max_length=255),
        size=8,
        default=list,
        help_text=(
            "A list of up to 8 URLs of images for this product, "
            "meant to be displayable to the customer."
        ),
    )
    shippable = models.BooleanField(
        null=True, help_text="Whether this product is shipped (i.e., physical goods)."
    )
    package_dimensions = models.JSONField(
        null=True,
        help_text="The dimensions of this product for shipping purposes.",
    )
    url = models.URLField(
        max_length=500,
        null=True,
        help_text="A URL of a publicly-accessible webpage for this product.",
    )
    created = models.BigIntegerField(
        help_text=(
            "Time at which the object was created. "
            "Measured in seconds since the Unix epoch"
        )
    )
    updated = models.BigIntegerField(
        help_text=(
            "Time at which the object was last updated. "
            "Measured in seconds since the Unix epoch"
        )
    )

    class Meta:
        abstract = True



# Third Party Stuff
from django.contrib.postgres.fields import ArrayField
from django.db import models

# Django Stripe Stuff
from django_stripe.models.abstracts.mixins import AbstractStripeModel
from django_stripe.utils import Currency


class AbstractStripeCustomer(AbstractStripeModel):
    """
    Customer objects allow us to perform recurring charges and track multiple
    charges that are associated with the same customer
    Stripe documentation: https://stripe.com/docs/api/customers
    """

    EXEMPT = "exempt"
    REVERSE = "reverse"
    NONE = "none"

    TAX_EXEMPT_TYPES = ((EXEMPT, "Exempt"), (REVERSE, "Reverse"), (NONE, "None"))

    # contact details
    name = models.TextField(
        max_length=255,
        blank=True,
        help_text="The customer's full name or business name",
    )
    description = models.TextField(
        max_length=255,
        blank=True,
        help_text=(
            "An arbitrary string attached to the object. "
            "Often useful for displaying to users."
        ),
    )
    email = models.EmailField(blank=True, db_index=True)
    address = models.JSONField(
        null=True, blank=True, help_text="The customer's address"
    )

    balance = models.DecimalField(
        decimal_places=2,
        max_digits=9,
        null=True,
        blank=True,
        help_text=(
            "Current balance (in cents), if any, being stored on the customer's "
            "account. "
            "If negative, the customer has credit to apply to the next invoice. "
            "If positive, the customer has an amount owed that will be added to the "
            "next invoice. The balance does not refer to any unpaid invoices; it "
            "solely takes into account amounts that have yet to be successfully "
            "applied to any invoice. This balance is only taken into account for "
            "recurring billing purposes (i.e., subscriptions, invoices, invoice items)"
        ),
    )
    currency = models.CharField(
        choices=Currency.choices,
        default=Currency.USD.value,
        max_length=3,
        help_text=(
            "The currency the customer can be charged "
            "in for recurring billing purposes"
        ),
    )
    delinquent = models.BooleanField(
        default=False,
        help_text=(
            "Whether or not the latest charge for the "
            "customer's latest invoice has failed"
        ),
    )
    default_source = models.TextField(blank=True)
    shipping = models.JSONField(
        null=True,
        blank=True,
        help_text="Shipping information associated with the customer",
    )
    tax_exempt = models.CharField(
        choices=TAX_EXEMPT_TYPES,
        max_length=16,
        default=NONE,
        help_text="Describes the customer's tax exemption status. When set to reverse, "
        'invoice and receipt PDFs include the text "Reverse charge"',
    )
    preferred_locales = ArrayField(
        models.CharField(default="", blank=True, max_length=255),
        default=list,
        help_text=(
            "The customer's preferred locales (languages), ordered by preference"
        ),
    )

    invoice_prefix = models.CharField(
        default="",
        blank=True,
        max_length=255,
        help_text=(
            "The prefix for the customer used to generate unique invoice numbers"
        ),
    )
    invoice_settings = models.JSONField(
        null=True, blank=True, help_text="The customer's default invoice settings"
    )
    is_active = models.BooleanField(default=True)

    class Meta:
        abstract = True



# Third Party Stuff
import stripe
from django.apps import apps
from django.conf import settings

# Django Stripe Stuff
from django_stripe.actions.mixins import (
    StripeSoftDeleteActionMixin,
    StripeSyncActionMixin,
)
from django_stripe.models import StripeCustomer


class StripeCustomerAction(StripeSyncActionMixin, StripeSoftDeleteActionMixin):
    """
    Synchronizes a local StripeCustomer data from the Stripe API
    Syncing is done by retrieving a batch of customers from the Stripe API
    and then iterating over them and calling sync method on each of them.

    Example:
        from django_stripe.actions import StripeCustomerAction
        stripe_action = StripeCustomerAction()
        stripe_action.sync_all()
    """

    model_class = StripeCustomer
    stripe_object_class = stripe.Customer

    def post_set_default(self, defaults: dict):
        """
        Sets default values for stripe data
        Args:
            defaults: defaults data
        """
        user_model_class = apps.get_model(settings.AUTH_USER_MODEL)
        user = user_model_class.objects.filter(email=defaults["email"]).first()
        defaults["user"] = user



# Third Party Stuff
from django.db import models

# Django Stripe Stuff
from django_stripe.models.abstracts import AbstractStripePrice


class StripePrice(AbstractStripePrice):
    product = models.ForeignKey(
        "django_stripe.StripeProduct",
        to_field="stripe_id",
        on_delete=models.CASCADE,
        null=True,
    )



# Third Party Stuff
from django.db import models

# Django Stripe Stuff
from django_stripe.models.abstracts.mixins import AbstractStripeModel


class AbstractStripeEvent(AbstractStripeModel):
    kind = models.CharField(max_length=255)
    webhook_message = models.JSONField()
    validated_message = models.JSONField(null=True, blank=True)
    valid = models.BooleanField(null=True)
    processed = models.BooleanField(default=False)
    request = models.JSONField(
        null=True,
        help_text=(
            "Information on the API request that instigated the event, "
            "If null, the event was automatic"
            " (e.g., Stripe’s automatic subscription handling)"
        ),
    )
    pending_webhooks = models.PositiveIntegerField(
        default=0,
        help_text=(
            "Number of webhooks that have yet to be successfully "
            "delivered (i.e., to return a 20x response) "
            "to the URLs we’ve specified"
        ),
    )
    api_version = models.CharField(max_length=128, blank=True)

    @property
    def message(self):
        return self.validated_message

    def __str__(self):
        return "{} - {}".format(self.kind, self.stripe_id)

    class Meta:
        abstract = True



# Standard Library
import logging

# Third Party Stuff
import stripe
from django.core.management import BaseCommand

# Django Stripe Stuff
from django_stripe.actions import StripeProduct

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Sync (UPDATE_OR_CREATE in local DB) products from stripe

    command: python manage.py sync_stripe_products
    """

    help = "Sync products"

    def handle(self, *args, **options):
        if not stripe.api_key:
            logger.info("Stripe API key not set while syncing products")
            return

        StripeProduct.sync_all()
        logger.info("Synced stripe products")



from types import DynamicClassAttribute

from django.db.models import TextChoices


class Currency(TextChoices):
    USD = "usd"
    AED = "aed"
    AFN = "afn"
    ALL = "all"
    AMD = "amd"
    ANG = "ang"
    AOA = "aoa"
    ARS = "ars"
    AUD = "aud"
    AWG = "awg"
    AZN = "azn"
    BAM = "bam"
    BBD = "bbd"
    BDT = "bdt"
    BGN = "bgn"
    BIF = "bif"
    BMD = "bmd"
    BND = "bnd"
    BOB = "bob"
    BRL = "brl"
    BSD = "bsd"
    BWP = "bwp"
    BYN = "byn"
    BZD = "bzd"
    CAD = "cad"
    CDF = "cdf"
    CHF = "chf"
    CLP = "clp"
    CNY = "cny"
    COP = "cop"
    CRC = "crc"
    CVE = "cve"
    CZK = "czk"
    DJF = "djf"
    DKK = "dkk"
    DOP = "dop"
    DZD = "dzd"
    EGP = "egp"
    ETB = "etb"
    EUR = "eur"
    FJD = "fjd"
    FKP = "fkp"
    GBP = "gbp"
    GEL = "gel"
    GIP = "gip"
    GMD = "gmd"
    GNF = "gnf"
    GTQ = "gtq"
    GYD = "gyd"
    HKD = "hkd"
    HNL = "hnl"
    HTG = "htg"
    HUF = "huf"
    IDR = "idr"
    ILS = "ils"
    INR = "inr"
    ISK = "isk"
    JMD = "jmd"
    JPY = "jpy"
    KES = "kes"
    KGS = "kgs"
    KHR = "khr"
    KMF = "kmf"
    KRW = "krw"
    KYD = "kyd"
    KZT = "kzt"
    LAK = "lak"
    LBP = "lbp"
    LKR = "lkr"
    LRD = "lrd"
    LSL = "lsl"
    MAD = "mad"
    MDL = "mdl"
    MGA = "mga"
    MKD = "mkd"
    MMK = "mmk"
    MNT = "mnt"
    MOP = "mop"
    MUR = "mur"
    MVR = "mvr"
    MWK = "mwk"
    MXN = "mxn"
    MYR = "myr"
    MZN = "mzn"
    NAD = "nad"
    NGN = "ngn"
    NIO = "nio"
    NOK = "nok"
    NPR = "npr"
    NZD = "nzd"
    PAB = "pab"
    PEN = "pen"
    PGK = "pgk"
    PHP = "php"
    PKR = "pkr"
    PLN = "pln"
    PYG = "pyg"
    QAR = "qar"
    RON = "ron"
    RSD = "rsd"
    RUB = "rub"
    RWF = "rwf"
    SAR = "sar"
    SBD = "sbd"
    SCR = "scr"
    SEK = "sek"
    SGD = "sgd"
    SHP = "shp"
    SLE = "sle"
    SOS = "sos"
    SRD = "srd"
    STD = "std"
    SZL = "szl"
    THB = "thb"
    TJS = "tjs"
    TOP = "top"
    TRY = "try"
    TTD = "ttd"
    TWD = "twd"
    TZS = "tzs"
    UAH = "uah"
    UGX = "ugx"
    UYU = "uyu"
    UZS = "uzs"
    VND = "vnd"
    VUV = "vuv"
    WST = "wst"
    XAF = "xaf"
    XCD = "xcd"
    XOF = "xof"
    XPF = "xpf"
    YER = "yer"
    ZAR = "zar"
    ZMW = "zmw"

    @DynamicClassAttribute
    def label(self):
        return self._label_.upper()



# Third Party Stuff
from django.conf import settings
from django.db import models

# Django Stripe Stuff
from django_stripe.models.abstracts.core import AbstractStripeCustomer


class StripeCustomer(AbstractStripeCustomer):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        related_name="stripe_customers",
    )



# Third Party Stuff
import stripe

# Django Stripe Stuff
from django_stripe.actions.core import StripeCustomer
from django_stripe.actions.mixins import (
    StripeSoftDeleteActionMixin,
    StripeSyncActionMixin,
)
from django_stripe.models import StripeSubscription


class StripeSubscriptionAction(StripeSyncActionMixin, StripeSoftDeleteActionMixin):
    """
    Synchronizes a local StripeSubscription data from the Stripe API

    Syncing is done by retrieving a batch of subscriptions from the Stripe API
    and then iterating over them and calling sync method on each of them.

    Example:
        from django_stripe.actions import StripeSubscriptionAction
        stripe_action = StripeSubscriptionAction()
        stripe_action.sync_all()
    """

    model_class = StripeSubscription
    stripe_object_class = stripe.Subscription

    def pre_set_defualt(self, stripe_data: dict):
        """
        Add customer data to stripe_data
        """
        stripe_data["customer"] = StripeCustomer.objects.get(
            stripe_id=stripe_data["customer"]
        )



# Django Stripe Stuff
from django_stripe.actions import StripePriceAction
from django_stripe.webhooks.register import StripeWebhook


class PriceStripeWebhook(StripeWebhook):
    def process_webhook(self):
        if self.event.validated_message:
            StripePriceAction().sync(
                self.event.validated_message["data"]["object"].copy()
            )


class PriceCreatedWebhook(PriceStripeWebhook):
    name = "price.created"
    description = "Occurs whenever a new price is created."


class PriceUpdatedWebhook(PriceStripeWebhook):
    name = "price.updated"
    description = "Occurs whenever any property of a price changes."


class PriceDeletedWebhook(PriceStripeWebhook):
    name = "price.deleted"
    description = "Occurs whenever a price is deleted."

    def process_webhook(self):
        StripePriceAction().soft_delete(
            self.event.validated_message["data"]["object"]["id"]
        )



# Standard Library
import json

# Third Party Stuff
import stripe
from six import with_metaclass

# Django Stripe Stuff
from django_stripe.actions import StripeEventAction
from django_stripe.webhooks.webhooks import WebhookRegistry

registry = WebhookRegistry()
del WebhookRegistry


class Registerable(type):
    def __new__(cls, clsname, bases, attrs):
        new_class = super(Registerable, cls).__new__(cls, clsname, bases, attrs)

        if getattr(new_class, "name", None) is not None:
            registry.register(new_class)

        return new_class


class StripeWebhook(with_metaclass(Registerable, object)):
    """
    REGISTRY: webhook registry
    name: webhook event name
    """

    REGISTRY = registry
    name = None

    def __init__(self, event):
        if event.kind != self.name:
            raise Exception(
                "The Webhook handler ({}) received the wrong type of Event ({})".format(
                    self.name, event.kind
                )
            )
        self.event = event

    def validate(self):
        """
        Validate incoming events
        We fetch the event data to ensure it is legit
        """
        evt = stripe.Event.retrieve(
            self.event.stripe_id,
        )
        self.event.validated_message = json.loads(
            json.dumps(
                dict(evt),
                sort_keys=True,
            )
        )
        self.event.valid = self.is_event_valid(
            self.event.webhook_message["data"], self.event.validated_message["data"]
        )
        self.event.save()

    @staticmethod
    def is_event_valid(webhook_message_data, validated_message_data):
        return (
            "object" in webhook_message_data
            and "object" in validated_message_data
            and webhook_message_data["object"]["id"]
            == validated_message_data["object"]["id"]
        )

    def send_signal(self):
        signal = self.REGISTRY.get_signal(self.name)
        if signal:
            return signal.send(sender=self.__class__, event=self.event)

    def process(self):
        if self.event.processed:
            return
        self.validate()
        if not self.event.valid:
            return

        try:
            StripeEventAction().link_customer(self.event)
            self.process_webhook()
            self.send_signal()
            self.event.processed = True
            self.event.save()
        except Exception as e:
            raise e

    def process_webhook(self):
        return



from django_stripe.models.products.coupons import StripeCoupon
from django_stripe.models.products.prices import StripePrice
from django_stripe.models.products.products import StripeProduct

__all__ = ("StripeProduct", "StripePrice", "StripeCoupon")



# Third Party Stuff
from django.apps import apps
from django.conf import settings

# Django Stripe Stuff
from django_stripe.actions import StripeCustomerAction
from django_stripe.models import StripeCustomer
from django_stripe.webhooks.register import StripeWebhook


class CustomerUpdatedWebhook(StripeWebhook):
    name = "customer.updated"
    description = "Occurs whenever any property of a customer changes."

    def process_webhook(self):
        if self.event.customer:
            stripe_customer = self.event.message["data"]["object"].copy()
            StripeCustomerAction().sync(stripe_customer)


class CustomerCreatedWebhook(StripeWebhook):
    name = "customer.created"
    description = "Occurs whenever a new customer is created."

    def process_webhook(self):
        stripe_customer = self.event.message["data"]["object"]
        email = stripe_customer["email"]
        stripe_id = self.event.message["data"]["object"]["id"]
        User = apps.get_model(settings.AUTH_USER_MODEL)
        user = User.objects.filter(email=email).first()

        if user and not user.stripe_customers.exists():
            # create customer
            data = {
                "user": user,
                "email": user.email,
                "is_active": True,
                "defaults": {"stripe_id": stripe_id},
            }
            customer, _ = StripeCustomer.objects.get_or_create(**data)

            # link customer to event
            self.event.customer = customer
            self.event.save()

            # sync customer
            StripeCustomerAction().sync(stripe_customer)


class CustomerDeletedWebhook(StripeWebhook):
    name = "customer.deleted"
    description = "Occurs whenever a customer is deleted."

    def process_webhook(self):
        if self.event.customer:
            StripeCustomerAction().soft_delete(self.event.customer.stripe_id)



# Third Party Stuff
from django.contrib.admin import AdminSite
from django.db.models.base import ModelBase

# Django Stripe Stuff
from django_stripe.settings import stripe_settings


class StripeAdminSite(AdminSite):
    def register(self, model_or_iterable, admin_class=None, **options):
        if "__all__" in stripe_settings.ALLOWED_ADMIN_MODELS:
            super().register(model_or_iterable, admin_class, **options)

        if isinstance(model_or_iterable, ModelBase):
            model_or_iterable = [model_or_iterable]

        # Filter models that are allowed to be registered
        allowed_model_or_iterable = []
        for model in model_or_iterable:
            if model._meta.model_name in stripe_settings.ALLOWED_ADMIN_MODELS:
                allowed_model_or_iterable.append(model)

        super().register(allowed_model_or_iterable)



from django_stripe.webhooks.payment_methods.cards import (
    CustomerCardCreatedWebhook,
    CustomerCardDeletedWebhook,
    CustomerCardUpdatedWebhook,
)

__all__ = (
    "CustomerCardCreatedWebhook",
    "CustomerCardUpdatedWebhook",
    "CustomerCardDeletedWebhook",
)



# Third Party Stuff
from django.core.exceptions import ObjectDoesNotExist
from django.http import Http404

# Django Stripe Stuff
from django_stripe.models import StripeCustomer, StripeEvent


class StripeEventAction:
    @classmethod
    def add(
        cls,
        stripe_id,
        kind,
        livemode,
        api_version,
        message,
        request=None,
        pending_webhooks=0,
    ):
        """
        Adds and processes an event from a received webhook
        Args:
            stripe_id: the stripe id of the event
            kind: the label of the event
            livemode: True or False if the webhook was sent from livemode or not
            message: the data of the webhook
            request_id: the id of the request that initiated the webhook
            pending_webhooks: the number of pending webhooks
        """
        event = StripeEvent.objects.create(
            stripe_id=stripe_id,
            kind=kind,
            livemode=livemode,
            webhook_message=message,
            api_version=api_version,
            request=request,
            pending_webhooks=pending_webhooks,
        )

        # Django Stripe Stuff
        from django_stripe.webhooks.register import registry

        WebhookClass = registry.get(kind)
        if WebhookClass is not None:
            webhook = WebhookClass(event)
            webhook.process()

    def link_customer(self, event):
        """
        Links a customer referenced in a webhook event message to the event object
        Args:
            event: the django_stripe.stripe.models.Event object to link
        """

        if event.kind == "customer.created":
            return

        customer_crud_events = [
            "customer.updated",
            "customer.deleted",
        ]
        event_data_object = event.message["data"]["object"]
        if event.kind in customer_crud_events:
            stripe_customer_id = event_data_object["id"]
        else:
            stripe_customer_id = event_data_object.get("customer", None)

        if stripe_customer_id is not None:
            try:
                customer = StripeCustomer.objects.get(stripe_id=stripe_customer_id)
            except ObjectDoesNotExist:
                raise Http404(
                    f"Stripe customer does not exist for event={event.stripe_id}"
                )

            event.customer = customer
            event.save()

        return event



# Third Party Stuff
from django.db import models

# Django Stripe Stuff
from django_stripe.models.abstracts.mixins import AbstractStripeModel
from django_stripe.utils import Currency


class AbstractStripePrice(AbstractStripeModel):
    """
    Prices define the unit cost, currency, and (optional) billing cycle for both
    recurring and one-time purchases of products. Products help you track inventory
    or provisioning, and prices help you track payment terms. Different physical
    goods or levels of service should be represented by products, and pricing options
    should be represented by prices. This approach lets you change prices without
    having to change your provisioning scheme.

    For example,
    you might have a single "gold" product that has
    prices for $10/month, $100/year, and €9 once.
    Stripe documentation: https://stripe.com/docs/api/prices
    """

    ONE_TIME = "one_time"
    RECURRING = "recurring"

    PRICE_TYPES = [
        (ONE_TIME, "One Time"),
        (RECURRING, "Recurring"),
    ]

    TAX_INCLUSIVE = "inclusive"
    TAX_EXCLUSIVE = "exclusive"
    TAX_UNSPECIFIED = "unspecified"

    TAX_BEHAVIOR_TYPES = [
        (TAX_INCLUSIVE, "TAX Inclusive"),
        (TAX_EXCLUSIVE, "TAX Exclusive"),
        (TAX_INCLUSIVE, "TAX Unspecified"),
    ]

    PER_UNIT = "per_unit"
    TIERED = "tiered"

    BILLING_SCHEME_TYPES = [
        (PER_UNIT, "Per Unit"),
        (TIERED, "Tiered"),
    ]

    active = models.BooleanField(
        help_text="Whether the price can be used for new purchases."
    )
    currency = models.CharField(
        choices=Currency.choices,
        default=Currency.USD.value,
        max_length=3,
        help_text=(
            "Three-letter ISO currency code, in lowercase. "
            "Must be a supported currency."
        ),
    )
    nickname = models.CharField(
        max_length=255,
        null=True,
        help_text="A brief description of the price, hidden from customers.",
    )
    recurring = models.JSONField(
        null=True,
        help_text=(
            "The recurring components of a price such as interval and usage_type."
        ),
    )

    type = models.CharField(
        choices=PRICE_TYPES,
        max_length=16,
        help_text=(
            "One of one_time or recurring depending on whether the price is for "
            "a one-time purchase or a recurring (subscription) purchase."
        ),
    )
    custom_unit_amount = models.JSONField(
        null=True,
        help_text=(
            "When set, provides configuration for the amount to be adjusted "
            "by the customer during Checkout Sessions and Payment Links."
        ),
    )
    unit_amount = models.BigIntegerField(
        null=True,
        help_text=(
            "The unit amount in cents to be charged, represented as a whole "
            "integer if possible. Null if a sub-cent precision is required"
        ),
    )
    unit_amount_decimal = models.DecimalField(
        null=True,
        max_digits=19,
        decimal_places=12,
        help_text=(
            "The unit amount in cents to be charged, represented as a decimal "
            "string with at most 12 decimal places"
        ),
    )
    billing_scheme = models.CharField(
        choices=BILLING_SCHEME_TYPES,
        max_length=16,
        help_text=(
            "Describes how to compute the price per period. Either per_unit or tiered."
            "per_unit indicates that the fixed amount "
            "(specified in unit_amount or unit_amount_decimal)"
            "will be charged per unit in quantity "
            "(for prices with usage_type=licensed),"
            "or per unit of total usage (for prices with usage_type=metered). "
            "Tiered indicates that the unit"
            "pricing will be computed using a tiering strategy "
            "as defined using the tiers and tiers_mode attributes."
        ),
    )

    tax_behavior = models.CharField(
        choices=TAX_BEHAVIOR_TYPES,
        max_length=16,
        help_text=(
            "Specifies whether the price is considered "
            "inclusive of taxes or exclusive of taxes."
            "One of inclusive, exclusive, or unspecified. "
            "Once specified as either inclusive or exclusive, it cannot be changed."
        ),
    )
    tiers = models.JSONField(
        null=True,
        help_text=(
            "Each element represents a pricing tier. "
            "This parameter requires billing_scheme to be set to tiered."
            "See also the documentation for billing_scheme. "
            "This field is not included by default."
            "To include it in the response, expand the tiers field."
        ),
    )
    tiers_mode = models.CharField(
        null=True,
        max_length=32,
        help_text=(
            "Defines if the tiering price should be graduated or volume based."
            "In volume-based tiering, the maximum quantity "
            "within a period determines the per unit price."
            "In graduated tiering, pricing can change as the quantity grows."
        ),
    )
    transform_quantity = models.JSONField(
        null=True,
        help_text=(
            "Apply a transformation to the reported usage or "
            "set quantity before computing the amount billed. "
            "Cannot be combined with tiers."
        ),
    )
    lookup_key = models.CharField(
        null=True,
        max_length=255,
        help_text=(
            "A lookup key used to retrieve prices dynamically from a static string. "
            "This may be up to 200 characters."
        ),
    )
    created = models.BigIntegerField(
        help_text=(
            "Time at which the object was created."
            "Measured in seconds since the Unix epoch"
        )
    )

    class Meta:
        abstract = True



from django_stripe.models.abstracts.core.customers import AbstractStripeCustomer
from django_stripe.models.abstracts.core.events import AbstractStripeEvent

__all__ = ("AbstractStripeCustomer", "AbstractStripeEvent")



# Django Stripe Stuff
from django_stripe.actions import StripeCouponAction
from django_stripe.webhooks.register import StripeWebhook


class CouponStripeWebhook(StripeWebhook):
    def process_webhook(self):
        if self.event.validated_message:
            StripeCouponAction().sync(
                self.event.validated_message["data"]["object"].copy()
            )


class CouponCreatedWebhook(CouponStripeWebhook):
    name = "coupon.created"
    description = "Occurs whenever a new coupon is created."


class CouponUpdatedWebhook(CouponStripeWebhook):
    name = "coupon.updated"
    description = "Occurs whenever any property of a coupon changes."


class CouponDeletedWebhook(CouponStripeWebhook):
    name = "coupon.deleted"
    description = "Occurs whenever a coupon is deleted."

    def process_webhook(self):
        StripeCouponAction().soft_delete(
            self.event.validated_message["data"]["object"]["id"]
        )



# Standard Library
import logging

# Third Party Stuff
import stripe
from django.apps import apps
from django.conf import settings
from django.core.management import BaseCommand
from stripe.error import InvalidRequestError

# Django Stripe Stuff
from django_stripe.actions import StripeCustomer

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Sync
    (ONLY UPDATE, it doesn't create customers if not exist in local DB)
    customers from stripe

    command: python manage.py sync_stripe_customers
    """

    help = "Sync customers data"

    def handle(self, *args, **options):
        if not stripe.api_key:
            logger.info("Stripe API key not set")
            return

        user_model_class = apps.get_model(settings.AUTH_USER_MODEL)
        users = user_model_class.objects.all()
        total = users.count()
        count = 0

        for user in users:
            customer = StripeCustomer.get(user)

            # show percentage of local customer's synced with stripe
            if customer:
                count += 1
                percent = int(round(100 * (float(count) / float(total))))
                self.stdout.write(
                    "[{0}/{1} {2}%] Syncing {3} {4}\n".format(
                        count,
                        total,
                        percent,
                        user.first_name,
                        user.last_name,
                    )
                )

                # sync local customer with stripe
                try:
                    customer = StripeCustomer.sync(customer)
                except InvalidRequestError as exc:
                    if exc.http_status == 404:
                        # This user doesn't exist (might be in test mode)
                        logger.info(
                            "Stripe customer doesn't exist, user_id=%s, customer_id=%s",
                            user.id,
                            customer.stripe_id,
                        )
                        continue
                    raise exc



# Third Party Stuff
from django.db import models

# Django Stripe Stuff
from django_stripe.models.abstracts.mixins import AbstractStripeModel


class AbstractStripeCard(AbstractStripeModel):
    """
    We can store multiple cards on a customer in order to charge the customer later.
    We can also store multiple debit cards
    on a recipient in order to transfer to those cards later.
    Stripe documentation: https://stripe.com/docs/api/cards
    """

    name = models.TextField(null=True, blank=True)
    address_line_1 = models.TextField(null=True, blank=True)
    address_line_1_check = models.CharField(null=True, blank=True, max_length=64)
    address_line_2 = models.TextField(null=True, blank=True)
    address_city = models.TextField(null=True, blank=True)
    address_state = models.TextField(null=True, blank=True)
    address_country = models.TextField(null=True, blank=True)
    address_zip = models.TextField(null=True, blank=True)
    address_zip_check = models.CharField(null=True, blank=True, max_length=64)
    brand = models.TextField(null=True, blank=True)
    country = models.CharField(null=True, blank=True, max_length=2)
    cvc_check = models.CharField(max_length=32, blank=True, null=True)
    dynamic_last4 = models.CharField(max_length=4, blank=True, null=True)
    tokenization_method = models.CharField(max_length=32, blank=True, null=True)
    exp_month = models.IntegerField()
    exp_year = models.IntegerField()
    funding = models.CharField(max_length=15, blank=True, null=True)
    last4 = models.CharField(max_length=4, blank=True, null=True)
    fingerprint = models.TextField(blank=True, null=True)

    def __repr__(self):
        return "Card(pk={!r}, customer={!r})".format(
            self.pk,
            getattr(self, "customer", None),
        )

    class Meta:
        abstract = True



from django_stripe.webhooks.core.customers import (
    CustomerCreatedWebhook,
    CustomerDeletedWebhook,
    CustomerUpdatedWebhook,
)

__all__ = ("CustomerCreatedWebhook", "CustomerUpdatedWebhook", "CustomerDeletedWebhook")



# Third Party Stuff
from django.dispatch import Signal


class WebhookRegistry(object):
    """
    WebhookRegistry is a class that registers webhooks.

    Webhooks are real-time notifications that are sent to your application
    when events happen in your Stripe account. For example, when a subscription
    is created, a webhook is sent to your application. You can then use this webhook
    to run actions in your application.

    WebhookRegistry is designed to be used with django. It provides a simple way to
    register webhook events in your application.

    Example:
        from django_stripe.webhooks import WebhookRegistry

        webhook_registry = WebhookRegistry()
        webhook_registry.register(PaymentSucceededWebhook)

        # process webhook events
        webhook_registry.process_webhook(
            {"id": "evt_123456789", "type": "payment_succeeded"}
        )

        # process webhook events from stripe
        webhook_registry.process_webhook(stripe.Event.retrieve("evt_123456789"))
    """

    def __init__(self):
        self._registry = {}

    def register(self, webhook):
        self._registry[webhook.name] = {
            "webhook": webhook,
            "signal": Signal(providing_args=["event"]),
        }

    def keys(self):
        return self._registry.keys()

    def get(self, name, default=None):
        try:
            return self[name]["webhook"]
        except KeyError:
            return default

    def get_signal(self, name, default=None):
        try:
            return self[name]["signal"]
        except KeyError:
            return default

    def signals(self):
        return {key: self.get_signal(key) for key in self.keys()}

    def __getitem__(self, name):
        return self._registry[name]



# Django Stripe Stuff
from django_stripe.actions import StripeCardAction
from django_stripe.webhooks.register import StripeWebhook


class CustomerCardStripeWebhook(StripeWebhook):
    def process_webhook(self):
        StripeCardAction().sync(
            self.event.customer, self.event.validated_message["data"]["object"]
        )


class CustomerCardCreatedWebhook(CustomerCardStripeWebhook):
    name = "customer.source.created"
    description = "Occurs whenever a new source is created for the customer."


class CustomerCardDeletedWebhook(StripeWebhook):
    name = "customer.source.deleted"
    description = "Occurs whenever a source is removed from a customer."

    def process_webhook(self):
        StripeCardAction().delete(self.event.validated_message["data"]["object"]["id"])


class CustomerCardUpdatedWebhook(CustomerCardStripeWebhook):
    name = "customer.source.updated"
    description = "Occurs whenever a source's details are changed."



# Django Stripe Stuff
from django_stripe.actions.billings import StripeSubscriptionAction
from django_stripe.actions.core import StripeCustomerAction
from django_stripe.actions.events import StripeEventAction
from django_stripe.actions.payment_methods import StripeCardAction
from django_stripe.actions.products import (
    StripeCouponAction,
    StripePriceAction,
    StripeProductAction,
)
from django_stripe.actions.webhooks import StripeWebhook

__all__ = [
    "StripeCustomerAction",
    "StripeEventAction",
    "StripeCardAction",
    "StripeCouponAction",
    "StripePriceAction",
    "StripeProductAction",
    "StripeSubscriptionAction",
    "StripeWebhook",
]



# Third Party Stuff
from django.conf import settings
from django.test.signals import setting_changed
from django.utils.module_loading import import_string

DEFAULTS = {
    "ALLOWED_MODELS": ["__all__"],
    "ALLOWED_ADMIN_MODELS": ["__all__"],
    "STRIPE_ROUTER": "django_stripe.routers.StripeRouter",
    "STRIPE_SITE": "django_stripe.sites.StripeAdminSite",
    "API_VERSION": "",
    "API_KEY": "",
}

IMPORT_STRINGS = []


def perform_import(val, setting_name):
    if val is None:
        return None
    elif isinstance(val, str):
        return import_from_string(val, setting_name)
    return val


def import_from_string(val, setting_name):
    try:
        return import_string(val)
    except ImportError as e:
        msg = "Could not import '%s' for Stripe setting '%s'. %s: %s." % (
            val,
            setting_name,
            e.__class__.__name__,
            e,
        )
        raise ImportError(msg)


class StripeSettings:
    def __init__(self, user_settings=None, defaults=None, import_strings=None):
        if user_settings:
            self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, "STRIPE_CONFIG", {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid Stripe setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")


stripe_settings = StripeSettings(None, DEFAULTS, IMPORT_STRINGS)


def reload_stripe_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == "STRIPE_CONFIG":
        stripe_settings.reload()
        settings.DATABASE_ROUTERS.append(stripe_settings.STRIPE_ROUTER)


setting_changed.connect(reload_stripe_settings)



# Third Party Stuff
import stripe
from django.apps import AppConfig

# Django Stripe Stuff
from django_stripe.settings import stripe_settings


class DjangoStripeConfig(AppConfig):
    name = "django_stripe"

    def ready(self):
        if stripe_settings.API_VERSION:
            stripe.api_version = stripe_settings.API_VERSION
        stripe.api_key = stripe_settings.API_KEY



# Standard Library
import importlib

importlib.import_module("django_stripe.webhooks.billings")
importlib.import_module("django_stripe.webhooks.core")
importlib.import_module("django_stripe.webhooks.payment_methods")
importlib.import_module("django_stripe.webhooks.products")



# Third Party Stuff
from django.utils.translation import gettext_lazy as _


class StripeException(Exception):
    default_detail = "Stripe error"

    def __init__(self, detail=None):
        self.detail = detail
        if not self.detail:
            self.detail = self.default_detail
        super().__init__(self.detail)


class StripeAuthException(StripeException):
    default_detail = _("Stripe auth error")



from django_stripe.webhooks.products.coupons import (
    CouponCreatedWebhook,
    CouponDeletedWebhook,
    CouponUpdatedWebhook,
)
from django_stripe.webhooks.products.prices import (
    PriceCreatedWebhook,
    PriceDeletedWebhook,
    PriceUpdatedWebhook,
)
from django_stripe.webhooks.products.products import (
    ProductCreatedWebhook,
    ProductDeletedWebhook,
    ProductUpdatedWebhook,
)

__all__ = [
    "ProductCreatedWebhook",
    "ProductUpdatedWebhook",
    "ProductDeletedWebhook",
    "PriceCreatedWebhook",
    "PriceUpdatedWebhook",
    "PriceDeletedWebhook",
    "CouponCreatedWebhook",
    "CouponUpdatedWebhook",
    "CouponDeletedWebhook",
]



# Standard Library
import logging

# Third Party Stuff
from django.utils.encoding import smart_str
from stripe.error import InvalidRequestError

# Django Stripe Stuff
from django_stripe.actions import StripeEventAction
from django_stripe.models import StripeEvent

logger = logging.getLogger(__name__)


class StripeWebhook:
    """
    StripeWebhook is a class that processes stripe webhooks.

    Stripe webhooks are real-time notifications that are sent to your application
    when events happen in your stripe account. For example, when a subscription
    is created, a webhook is sent to your application. You can then use this webhook
    to run actions in your application.

    StripeWebhook is designed to be used with django. It provides a simple way to
    process webhook events in your application.

    Example:
        from django_stripe.actions import StripeWebhook
        from django_stripe.webhooks import PaymentSucceededWebhook

        webhook = StripeWebhook()
        webhook.register(PaymentSucceededWebhook)

        # process webhook events
        webhook.process_webhook({"id": "evt_123456789", "type": "payment_succeeded"})

        # process webhook events from stripe
        webhook.process_webhook(stripe.Event.retrieve("evt_123456789"))
    """

    @classmethod
    def process_webhook(cls, event_data):
        event = StripeEvent.objects.filter(stripe_id=event_data["id"]).first()

        if event:
            logger.info(
                "Found duplicate stripe event record with event_id=%s", event.id
            )
            return

        try:
            # create an event and process webhook
            StripeEventAction.add(
                stripe_id=event_data["id"],
                kind=event_data["type"],
                livemode=event_data["livemode"],
                message=event_data,
                api_version=event_data["api_version"],
                request=event_data["request"],
                pending_webhooks=event_data["pending_webhooks"],
            )
        except InvalidRequestError as e:
            event_id = event_data["id"]
            logger.info(
                "Error occurred while processing stripe webhook, "
                f"event_id={event_id}, error={smart_str(e)}"
            )
        return



from django_stripe.models.billings import StripeSubscription
from django_stripe.models.core import StripeCustomer, StripeEvent
from django_stripe.models.payment_methods import StripeCard
from django_stripe.models.products import StripeCoupon, StripePrice, StripeProduct

__all__ = (
    "StripeSubscription",
    "StripeCustomer",
    "StripeEvent",
    "StripeCard",
    "StripeProduct",
    "StripePrice",
    "StripeCoupon",
)



# Django Stripe Stuff
from django_stripe.actions import StripeProductAction
from django_stripe.webhooks.register import StripeWebhook


class ProductStripeWebhook(StripeWebhook):
    def process_webhook(self):
        if self.event.validated_message:
            StripeProductAction().sync(
                self.event.validated_message["data"]["object"].copy()
            )


class ProductCreatedWebhook(ProductStripeWebhook):
    name = "product.created"
    description = "Occurs whenever a new product is created."


class ProductUpdatedWebhook(ProductStripeWebhook):
    name = "product.updated"
    description = "Occurs whenever any property of a product changes."


class ProductDeletedWebhook(StripeWebhook):
    name = "product.deleted"
    description = "Occurs whenever a product is deleted."

    def process_webhook(self):
        StripeProductAction().soft_delete(
            self.event.validated_message["data"]["object"]["id"]
        )



# Third Party Stuff
from django.utils import timezone

# Django Stripe Stuff
from django_stripe.utils import convert_epoch


class StripeSoftDeleteActionMixin:
    """
    A mixin class that provides a soft_delete method which allows to soft delete
    local stripe objects. This is useful when you want to mark a stripe object as
    deleted without actually deleting it in the local database.

    Example:
        from django_stripe.actions import StripeSoftDeleteActionMixin
        from django_stripe.models import StripeCustomer

        class StripeCustomerAction(StripeSoftDeleteActionMixin):
            model_class = StripeCustomer

        stripe_action = StripeCustomerAction()
        stripe_action.soft_delete("cus_123456789")
    """

    model_class = None

    def soft_delete(self, stripe_id: str):
        """
        Deletes the local stripe object
        Args:
            stripe_id: the Stripe ID of the stripe object
        """
        obj = self.model_class.objects.filter(stripe_id=stripe_id).first()
        if obj:
            obj.date_purged = timezone.now()
            obj.save()


class StripeSyncActionMixin:
    """
    A mixin class that provides a sync method that synchronizes a local
    data from the Stripe API

    Syncing is done by retrieving a batch of objects from the Stripe API
    and then iterating over them and calling sync method on each of them.

    Example:
        from django_stripe.actions import StripeSyncActionMixin
        from django_stripe.models import StripeCustomer

        class StripeCustomerAction(StripeSyncActionMixin):
            model_class = StripeCustomer

        stripe_action = StripeCustomerAction()
        stripe_action.sync_all()
    """

    model_class = None
    stripe_object_class = None
    batch_size = 1000

    def pre_set_defualt(self, stripe_data: dict):
        """
        Override this method to set values in stripe_data
        before setting the default.
        Or perform certain actions before setting the default.
        Args:
            stripe_data: data from Stripe API
        """
        pass

    def post_set_default(self, defaults: dict):
        """
        Override this method to perform actions after setting the default.
        """
        pass

    def set_default(self, stripe_data: dict):
        defaults = {}

        for field in self.model_class._meta.get_fields():
            if field.name not in stripe_data:
                continue

            field_type = field.get_internal_type()

            if field_type == "DateTimeField":
                defaults[field.name] = (
                    convert_epoch(stripe_data[field.name])
                    if stripe_data[field.name]
                    else None
                )
            elif field_type in ["CharField", "TextField"]:
                defaults[field.name] = stripe_data[field.name] or ""
            else:
                defaults[field.name] = stripe_data[field.name]

        return defaults

    def sync(self, stripe_data: dict):
        """
        Synchronizes a local data from the Stripe API
        Args:
            stripe_data: data from Stripe API
        """

        self.pre_set_defualt(stripe_data)
        stripe_id = stripe_data.pop("id")
        defaults = self.set_default(stripe_data)
        self.post_set_default(defaults)

        model_obj, _ = self.model_class.objects.update_or_create(
            stripe_id=stripe_id, defaults=defaults
        )

        return model_obj

    def sync_by_ids(self, ids):
        """
        Synchronizes a local data from the Stripe API
        Args:
            ids: list of ids
        """
        stripe_data = self.stripe_object_class.retrieve(ids)
        self.sync(stripe_data)

    def _update_model_objs(
        self, model_objs: list[object], stripe_id_obj_map: dict[str, dict]
    ):
        """
        Updates model objects
        Args:
            model_objs: list of model objects
            stripe_id_obj_map: dict of stripe id and stripe object data to be updated
        """
        if not model_objs:
            return

        for model_obj in model_objs:
            stripe_id = model_obj.stripe_id
            data = stripe_id_obj_map[stripe_id]

            self.pre_set_defualt(data)
            defaults = self.set_default(data)
            self.post_set_default(defaults)

            for key, value in defaults.items():
                setattr(model_obj, key, value)

            del stripe_id_obj_map[stripe_id]

        self.model_class.objects.bulk_update(model_objs, fields=list(defaults.keys()))

    def _create_model_objs(self, stripe_id_obj_map: dict[str, dict]):
        """
        Creates model objects
        Args:
            stripe_id_obj_map: dict of stripe id and stripe object data to be created
        """
        if not stripe_id_obj_map:
            return

        model_objs = []

        for stripe_id, data in stripe_id_obj_map.items():
            self.pre_set_defualt(data)
            defaults = self.set_default(data)
            defaults["stripe_id"] = stripe_id
            self.post_set_default(defaults)

            model_objs.append(self.model_class(**defaults))

        self.model_class.objects.bulk_create(model_objs)

    def sync_batch(self, batch: list[dict]):
        """
        Synchronizes a batch of data from the Stripe API
        Args:
            batch: list of data from Stripe API
        """
        stripe_id_obj_map = {}
        for data in batch:
            stripe_id = data.pop("id")
            stripe_id_obj_map[stripe_id] = data

        model_objs = self.model_class.objects.filter(
            stripe_id__in=stripe_id_obj_map.keys()
        )
        self._update_model_objs(model_objs, stripe_id_obj_map)
        self._create_model_objs(stripe_id_obj_map)

    def sync_all(self):
        """
        Synchronizes all data from the Stripe API
        """
        objects = self.stripe_object_class.auto_paging_iter()
        stripe_ids = []
        batch = []

        for i, obj in enumerate(objects):
            stripe_ids.append(obj["id"])
            batch.append(obj)
            if (i + 1) % self.batch_size == 0:
                self.sync_batch(batch)
                batch = []

        self.sync_batch(batch)

        # sync deleted objects
        self.model_class.objects.exclude(stripe_id__in=stripe_ids).update(
            date_purged=timezone.now()
        )



# Third Party Stuff
from django.dispatch import Signal


class WebhookRegistry(object):
    def __init__(self):
        self._registry = {}

    def register(self, webhook):
        self._registry[webhook.name] = {
            "webhook": webhook,
            "signal": Signal(),
        }

    def keys(self):
        return self._registry.keys()

    def get(self, name, default=None):
        try:
            return self[name]["webhook"]
        except KeyError:
            return default

    def get_signal(self, name, default=None):
        try:
            return self[name]["signal"]
        except KeyError:
            return default

    def signals(self):
        return {key: self.get_signal(key) for key in self.keys()}

    def __getitem__(self, name):
        return self._registry[name]
